import cv2
import numpy as np
from skimage import measure
from skimage.draw import polygon2mask
from holocontour.image.structure_forest import generate_mask
from holocontour.contour.toolsbox import contour_mask_union, filter_contours_by_intensity
from holocontour.image.visual import plot_segmentation_result
from holocontour.image.region_growing import region_grow
from holocontour.image.processing import apply_histogram_matching, find_darkest_point, get_sharpening_kernel, erode_mask


def process_mask(img_org,
                 img,
                 avg_thresh,
                 min_contour_area,
                 seed_thresh,
                 median,
                 erode_ksize,
                 convex_hull):

    seed = find_darkest_point(img_org, generate_mask(img, use_convex_hull=convex_hull) > 0)
    seg_mask = region_grow(img, seed)

    contours = measure.find_contours(seg_mask, 0.5)
    filtered_contours = [c for c in contours if len(c) > min_contour_area]

    while True:
        outside = np.where((img < seed_thresh) & ~seg_mask)
        if len(outside[0]) == 0:
            break
        seed = (outside[0][0], outside[1][0])
        new_mask = region_grow(img, seed)
        seg_mask |= new_mask
        new_contours = measure.find_contours(new_mask, 0.5)
        filtered_contours += [c for c in new_contours if len(c) > min_contour_area]

    valid_contours = filter_contours_by_intensity(
        filtered_contours, img_org, avg_thresh, median
    )

    union = contour_mask_union(valid_contours, img_org.shape)

    outer_first = max(measure.find_contours(generate_mask(img, use_convex_hull=convex_hull), 0.5), key=len)
    init_mask_poly = polygon2mask(img_org.shape[:2], outer_first)
    eroded_init_mask = erode_mask(init_mask_poly, kernel_size=erode_ksize)
    outer = max(measure.find_contours(eroded_init_mask, 0.5), key=len)

    init_mask_poly = polygon2mask(img_org.shape[:2], outer)
    union &= init_mask_poly

    final_mask = np.zeros_like(union, dtype=np.uint8)
    final_contours = []

    for contour in measure.find_contours(union, 0.5):
        if len(contour) > min_contour_area:
            mask = polygon2mask(img_org.shape[:2], contour).astype(np.uint8)
            final_mask += mask
            final_contours.append(contour)

    return final_mask, final_contours, outer


def find_contours(img_org,
                 avg_thresh=81,
                 max_attempts=5,
                 increase_avg=5,
                 min_contour_area=30,
                 seed_thresh=45,
                 save_plot=False,
                 median=False,
                 hist_match=False,
                 ref_path=None,
                 erode_ksize=3,
                 convex_hull=False,
                 keep_init_mask=False,
                 median_blur_ksize=5,
                 sharpening_alpha=0):

    img = img_org.copy()

    if hist_match and ref_path:
        img = apply_histogram_matching(img_org, ref_path)

    blur = cv2.GaussianBlur(img, (3, 3), 0)
    img = cv2.addWeighted(img, 1 + sharpening_alpha, blur, -sharpening_alpha, 0)

    if median_blur_ksize > 1:
        img = cv2.medianBlur(img, median_blur_ksize)

    init_mask = generate_mask(img, use_convex_hull=convex_hull)

    if np.count_nonzero(init_mask) == 0:
        print("[WARNING] Empty init_mask — skipping image.")
        return np.zeros_like(img_org, dtype=bool), img_org

    # Attempt up to max_attempts times with increasing avg_thresh
    attempt = 0

    while attempt < max_attempts:
        final_mask, final_contours, outer = process_mask(
            img_org,
            img,
            avg_thresh,
            min_contour_area,
            seed_thresh,
            median,
            erode_ksize,
            convex_hull
        )

        if np.count_nonzero(final_mask) > 0:
            # Success
            if save_plot:
                plot = plot_segmentation_result(img_org, outer, final_contours)
            else:
                plot = None

            return (polygon2mask(img_org.shape[:2], outer) if keep_init_mask else final_mask) > 0, plot

        else:
            attempt += 1
            avg_thresh += increase_avg
            print(f"[INFO] Final mask empty — increasing avg_thresh to {avg_thresh} (attempt {attempt}/{max_attempts})")

    print("[WARNING] All attempts failed — returning initial mask.")
    outer_contour_first = max(measure.find_contours(init_mask, 0.5), key=len)
    init_mask_poly = polygon2mask(img_org.shape[:2], outer_contour_first)
    eroded_init_mask = erode_mask(init_mask_poly, kernel_size=erode_ksize)
    outer_contour = max(measure.find_contours(eroded_init_mask, 0.5), key=len)

    final_mask = eroded_init_mask
    final_contours = [outer_contour]

    if save_plot:
        plot = plot_segmentation_result(img_org, outer_contour, final_contours)
    else:
        plot = None

    return final_mask > 0, plot



from pathlib import Path
import numpy as np
import cv2
from skimage import io
from skimage.color import rgb2gray
from skimage.filters import sobel
from skimage.feature import canny
from skimage.morphology import binary_closing, disk, remove_small_objects, convex_hull_image
from skimage.measure import label, regionprops
from scipy.ndimage import binary_fill_holes
import importlib.resources as pkg_resources
from holocontour import model


def structured_forest_edges(gray_img, model_path):
    rgb_img = cv2.cvtColor(gray_img, cv2.COLOR_GRAY2RGB)
    sed = cv2.ximgproc.createStructuredEdgeDetection(model_path)
    edges = sed.detectEdges(np.float32(rgb_img) / 255.0)
    return edges


def region_refining(binary_image, area_thresh=25):
    cleaned = remove_small_objects(binary_image.astype(bool), min_size=area_thresh)
    filled = binary_fill_holes(cleaned)
    return filled.astype(np.uint8)


def canny_edge(image):
    if image.ndim == 3:
        image = rgb2gray(image)
    edges = canny(image)
    closed = binary_closing(edges, disk(3))
    filled = binary_fill_holes(closed)
    return region_refining(filled)


def sobel_edge(image):
    if image.ndim == 3:
        image = rgb2gray(image)
    edges = sobel(image)
    binary = edges > 0.05
    closed = binary_closing(binary, disk(3))
    filled = binary_fill_holes(closed)
    return region_refining(filled)


def region_methods():
    return {
        'structuredForest': None,  # handled separately
        'cannyEdge': canny_edge,
        'sobelEdge': sobel_edge
    }


def reg_props(mask, name_props):
    labeled_mask = label(mask)
    all_props = regionprops(labeled_mask)
    results = []
    for region in all_props:
        props = {name: getattr(region, name, None) for name in name_props}
        results.append(props)
    return results


def particle_sizer(code_dir, data_dir, im_format, save_dir, method_key, model_path, name_props, use_convex_hull):
    mask_dir = Path(save_dir) / "masks"
    mask_dir.mkdir(parents=True, exist_ok=True)

    image_paths = list(Path(data_dir).rglob(f'*{im_format}'))
    print(f"Found {len(image_paths)} images")

    methods = region_methods()
    method_func = methods.get(method_key)

    for img_path in image_paths:
        image = io.imread(img_path)
        gray = image[:, :, 0]
        # gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY) if image.ndim == 3 else image

        if method_key == 'structuredForest':
            edges = structured_forest_edges(gray, model_path)
            binary = edges > 0.05
        elif method_func:
            binary = method_func(gray)
        else:
            raise ValueError(f"Unknown method: {method_key}")

        binary = remove_small_objects(binary, min_size=50)
        if use_convex_hull:
            binary = convex_hull_image(binary)

        mask_path = mask_dir / (img_path.stem + "_mask.png")
        io.imsave(mask_path, (binary * 255).astype(np.uint8))

        props_list = reg_props(binary, name_props)
        for props in props_list:
            print(f"{img_path.name} → {props}")


def generate_mask(img, use_convex_hull=False):
    with pkg_resources.path(model, "model.yml") as model_path:
        edges = structured_forest_edges(img, str(model_path))

    binary = edges > 0.05
    binary = remove_small_objects(binary, min_size=50)

    if use_convex_hull:
        binary = convex_hull_image(binary)

    return binary



import cv2
import numpy as np
from skimage.exposure import match_histograms


def normalize(mask, range_val=255):
    """Normalize an image to a given range."""
    return (range_val * (mask - np.min(mask)) / np.ptp(mask)).astype(np.uint8)


def find_darkest_point(image, mask):
    """Find the darkest pixel within a mask."""
    return np.unravel_index(np.argmin(image[mask]), image.shape)


def apply_histogram_matching(img, ref_path):
    """Match histogram of image to reference."""
    ref = cv2.imread(ref_path, 0)
    matched = match_histograms(img, ref)
    matched = np.clip(matched, img.min(), img.max())
    return matched.astype(img.dtype)


def get_sharpening_kernel(size: int):
    """
    Generate a sharpening kernel of given size (odd number).
    Example:
        size=3 -> [[0, -1, 0], [-1, 5, -1], [0, -1, 0]]
        size=5 -> 5x5 version with cross neighbors -1
    """
    if size % 2 == 0 or size < 3:
        raise ValueError("Kernel size must be an odd integer >= 3")

    kernel = np.zeros((size, size), dtype=np.float32)

    # Put -1 in cross neighbors
    center = size // 2
    for i in range(size):
        kernel[center, i] = -1   # horizontal line
        kernel[i, center] = -1   # vertical line

    # Set center value (sum of negatives * -1 + 1)
    kernel[center, center] = -np.sum(kernel)

    return kernel


def erode_mask(mask, kernel_size=3, iterations=1):
    """
    Erodes a binary mask using an elliptical kernel.

    Parameters
    ----------
    mask : np.ndarray
        Binary mask (0 and 255 values).
    kernel_size : int
        Size of the elliptical kernel (must be odd, e.g., 3, 5, 7).
    iterations : int
        How many times erosion is applied.

    Returns
    -------
    eroded : np.ndarray
        Eroded binary mask.
    """
    # Ensure kernel size is at least 1 and odd
    kernel_size = max(1, kernel_size)
    if kernel_size % 2 == 0:
        kernel_size += 1

    # Convert to uint8 0/255
    if mask.dtype == np.bool_:
        mask = mask.astype(np.uint8) * 255
    elif mask.max() <= 1:
        mask = (mask > 0).astype(np.uint8) * 255
    else:
        mask = mask.astype(np.uint8)

    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (kernel_size, kernel_size))
    eroded = cv2.erode(mask, kernel, iterations=iterations)

    return eroded > 0


from morphocut import Node, Output, ReturnOutputs
from holocontour.contour.contour_detection import find_contours

@ReturnOutputs
class HoloContourNode(Node):
    mask = Output("mask")
    plot = Output("plot")
    outputs = (mask, plot)

    def __init__(self, img, contour_params):
        super().__init__()
        self.img = img
        self.contour_params = contour_params

    def transform(self, img):
        return find_contours(img, **self.contour_params)



import numpy as np


def region_grow(image, seed, max_iter=20000, tolerance=25):
    """Simple region growing algorithm from a seed."""
    mask = np.zeros_like(image, dtype=bool)
    region_mean = image[seed]
    stack = [seed]
    iter_count = 0

    while stack and iter_count < max_iter:
        x, y = stack.pop()
        if not mask[x, y]:
            mask[x, y] = True
            for dx, dy in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
                nx, ny = x + dx, y + dy
                if (
                    0 <= nx < image.shape[0] and
                    0 <= ny < image.shape[1] and
                    not mask[nx, ny] and
                    abs(image[nx, ny] - region_mean) < tolerance
                ):
                    stack.append((nx, ny))
        iter_count += 1

    return mask


import numpy as np
from skimage import measure
from skimage.draw import polygon2mask


def contour_mask_union(contours, shape):
    """Create a union mask from multiple contours."""
    union_mask = np.zeros(shape, dtype=bool)
    for contour in contours:
        union_mask |= measure.grid_points_in_poly(shape, contour)
    return union_mask


def filter_contours_by_intensity(contours, image, threshold, use_median=False):
    """Filter contours by intensity threshold."""
    filtered = []
    for contour in contours:
        req_mask = polygon2mask(image.shape[:2], contour).astype(bool)
        region_values = image[req_mask]
        stat = np.median(region_values) if use_median else np.mean(region_values)

        # contour_coords = np.array(list(zip(contour[:, 0].astype(int), contour[:, 1].astype(int))))
        # stat = np.mean(image[contour_coords[:, 0], contour_coords[:, 1]])

        if stat <= threshold:
            filtered.append(contour)
    return filtered